unit FmtBcdNew;
{$MODE OBJFPC}
{$MODESWITCH ADVANCEDRECORDS}
{$INLINE ON}

interface

uses SysUtils, Variants;

{$DEFINE SKIP}

type
  FmtBCDStringType = ShortString;

  PBCD = ^TBCD;

  { TBCD }
  TBCD = packed record
  private
    const
      CMaxPrecision = 64;
    type
      TPrecision = 0..CMaxPrecision; // 0 - for BCD zero
      TNumberOfDecimalPlaces = 0..CMaxPrecision - 1; // 0 - for whole numbers
      // Two digit in Byte (00..99), high nibble first
      TData = array[0..CMaxPrecision div 2 - 1] of Byte;
    var
      // Binary compatible with old type
      FPrecision: TPrecision;
      FInfo: bitpacked record
        Places: TNumberOfDecimalPlaces;
        Unused: Boolean;
        Negative: Boolean;
      end;
      FData: TData;
    function GetSignSpecialPlaces: Byte; inline;
    procedure SetNegative(const AValue: Boolean); inline;
  public
    // Old type compatible fields
    property Precision: TPrecision read FPrecision;
    property SignSpecialPlaces: Byte read GetSignSpecialPlaces;
    property Fraction: TData read FData;
    // New properties
    property Places: TNumberOfDecimalPlaces read FInfo.Places;
    property Negative: Boolean read FInfo.Negative write SetNegative;
  end;

  EBCDException = class(Exception);
  EBCDOverflowException = class(EBCDException);
  EBCDNotImplementedException = class(EBCDException);

function BCDPrecision(const BCD: TBCD): Word; inline;
function BCDScale(const BCD: TBCD): Word; inline;
function IsBCDNegative(const BCD: TBCD): Boolean; inline;

procedure BCDNegate(var BCD: TBCD); inline;

{ Returns True if successful, False if Int Digits needed to be truncated }
function NormalizeBCD(const InBCD: TBCD; out OutBCD: TBCD;
  const MaxPrecision, MaxPlaces: Integer): Boolean;

{$IFNDEF SKIP}
{ TBCD variant creation utils }
procedure VarFmtBCDCreate(out Dest: Variant; const BCD: TBCD);
function VarFmtBCDCreate: Variant;
function VarFmtBCDCreate(const Value: FmtBCDStringType; Precision, Scale: Word): Variant;
{$ifndef FPUNONE}
  function VarFmtBCDCreate(const Value: ValReal; Precision: Word = 18; Scale: Word = 4): Variant;
{$endif}
function VarFmtBCDCreate(const BCD: TBCD): Variant;
function VarIsFmtBCD(const Value: Variant): Boolean;
function VarFmtBCD: TVarType;
{$ENDIF SKIP}

{ Convert string/Double/Integer to BCD struct }
function StrToBCD(const Value: FmtBCDStringType): TBCD; overload;
function StrToBCD(const Value: FmtBCDStringType; const Format: TFormatSettings): TBCD; overload;
function TryStrToBCD(const Value: FmtBCDStringType; out BCD: TBCD): Boolean; overload;
function TryStrToBCD(const Value: FmtBCDStringType; out BCD: TBCD; const Format: TFormatSettings): Boolean; overload;
{$IFNDEF SKIP}
{$ifndef FPUNONE}
  function DoubleToBCD(const Value: ValReal): TBCD; inline;
  procedure DoubleToBCD(const Value: ValReal; out Result: TBCD);
{$endif}
function IntegerToBCD(const Value: Int64): TBCD;
function VarToBCD(const Value: Variant): TBCD;
function CurrToBCD(const Value: Currency; out BCD: TBCD;
  Precision: Integer = 32; Decimals: Integer = 4): Boolean;

{ Convert BCD struct to string/Double/Integer }
function BCDToStr(const BCD: TBCD): FmtBCDStringType;
function BCDToStr(const BCD: TBCD; const Format: TFormatSettings): FmtBCDStringType;
{$ifndef FPUNONE}
  function BCDToDouble(const BCD: TBCD): ValReal;
{$endif}
function BCDToInteger(const BCD: TBCD; Truncate: Boolean = False): Int64;
function BCDToCurr(const BCD: TBCD; out OutResult: currency): Boolean;

{ Formatting BCD as string }
function BCDToStrF(const BCD: TBCD; Format: TFloatFormat;
  const Precision, Digits: Integer): FmtBCDStringType;
function FormatBCD(const Format: string; const BCD: TBCD): FmtBCDStringType;

{ returns -1 if BCD1 < BCD2, 0 if BCD1 = BCD2, 1 if BCD1 > BCD2 }
function BCDCompare(const BCD1, BCD2: TBCD): Integer;
function CurrToBCD(const Value: Currency): TBCD; inline;

{$ifdef SUPPORT_COMP}
  function CompToBCD(const Value: Comp): TBCD; inline;
  function BCDToComp(const BCD: TBCD): Comp; inline;
{$endif}

// Add
procedure BCDAdd(const BCD1, BCD2: TBCD; out Result: TBCD);
procedure BCDAdd(const BCD1: TBCD; const Int2: Int64; out Result: TBCD);
procedure BCDAdd(const Int1: Int64; const BCD2: TBCD; out Result: TBCD); inline;
{$ifndef FPUNONE}
  procedure BCDAdd(const BCD1: TBCD; const Double2: ValReal; out Result: TBCD); inline;
  procedure BCDAdd(const Double1: ValReal; const BCD2: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDAdd(const BCD1: TBCD; const Curr2: Currency; out Result: TBCD); inline;
procedure BCDAdd(const Curr1: Currency; const BCD2: TBCD; out Result: TBCD); inline;
{$ifdef SUPPORT_COMP}
  procedure BCDAdd(const BCD1: TBCD; const Comp2: Comp; out Result: TBCD); inline;
  procedure BCDAdd(const Comp1: Comp; const BCD2: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDAdd(const BCD1: TBCD; const Str2: FmtBCDStringType; out Result: TBCD); inline;
procedure BCDAdd(const Str1: FmtBCDStringType; const BCD2: TBCD; out Result: TBCD); inline;
procedure BCDAdd(const Str1, Str2: FmtBCDStringType; out Result: TBCD); inline;

// Subtract
procedure BCDSubtract(const BCD1, BCD2: TBCD; out Result: TBCD);
procedure BCDSubtract(const BCD1: TBCD; const Int2: Int64; out Result: TBCD);
procedure BCDSubtract(const Int1: Int64; const BCD2: TBCD; out Result: TBCD); inline;
{$ifndef FPUNONE}
  procedure BCDSubtract(const BCD1: TBCD; const Double2: ValReal; out Result: TBCD); inline;
  procedure BCDSubtract(const Double1: ValReal; const BCD2: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDSubtract(const BCD1: TBCD; const Curr2: Currency; out Result: TBCD); inline;
procedure BCDSubtract(const Curr1: Currency; const BCD2: TBCD; out Result: TBCD); inline;
{$ifdef SUPPORT_COMP}
  procedure BCDSubtract(const BCD1: TBCD; const Comp2: Comp; out Result: TBCD); inline;
  procedure BCDSubtract(const Comp1: Comp; const BCD2: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDSubtract(const BCD1: TBCD; const Str2: FmtBCDStringType; out Result: TBCD); inline;
procedure BCDSubtract(const Str1: FmtBCDStringType; const BCD2: TBCD; out Result: TBCD); inline;
procedure BCDSubtract(const Str1, Str2: FmtBCDStringType; out Result: TBCD); inline;

// Multiply
procedure BCDMultiply(const BCD1, BCD2: TBCD; out Result: TBCD);
{$ifndef FPUNONE}
  procedure BCDMultiply(const BCD1: TBCD; const Double2: ValReal; out Result: TBCD); inline;
{$endif}
procedure BCDMultiply(const BCD1: TBCD; const Str2: FmtBCDStringType; out Result: TBCD); inline;
procedure BCDMultiply(const Str1, Str2: FmtBCDStringType; out Result: TBCD); inline;
procedure BCDMultiply(const BCD1: TBCD; const Int2: Int64; out Result: TBCD);
procedure BCDMultiply(const Int1: Int64; const BCD2: TBCD; out Result: TBCD); inline;
{$ifndef FPUNONE}
  procedure BCDMultiply(const Double1: ValReal; const BCD2: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDMultiply(const BCD1: TBCD; const Curr2: Currency; out Result: TBCD); inline;
procedure BCDMultiply(const Curr1: Currency; const BCD2: TBCD; out Result: TBCD); inline;
{$ifdef SUPPORT_COMP}
  procedure BCDMultiply(const BCD1: TBCD; const Comp2: Comp; out Result: TBCD); inline;
  procedure BCDMultiply(const Comp1: Comp; const BCD2: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDMultiply(const Str1: FmtBCDStringType; const BCD2: TBCD; out Result: TBCD); inline;

// Divide
procedure BCDDivide(const Dividend, Divisor: TBCD; out Result: TBCD);
{$ifndef FPUNONE}
  procedure BCDDivide(const Dividend: TBCD; const Divisor: ValReal; out Result: TBCD); inline;
{$endif}
procedure BCDDivide(const Dividend: TBCD; const Divisor: FmtBCDStringType; out Result: TBCD); inline;
procedure BCDDivide(const Dividend, Divisor: FmtBCDStringType; out Result: TBCD); inline;
procedure BCDDivide(const Dividend: TBCD; const Divisor: Int64; out Result: TBCD); inline;
procedure BCDDivide(const Dividend: Int64; const Divisor: TBCD; out Result: TBCD); inline;
{$ifndef FPUNONE}
  procedure BCDDivide(const Dividend: ValReal; const Divisor: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDDivide(const Dividend: TBCD; const Divisor: Currency; out Result: TBCD); inline;
procedure BCDDivide(const Dividend: Currency; const Divisor: TBCD; out Result: TBCD); inline;
{$ifdef SUPPORT_COMP}
  procedure BCDDivide(const Dividend: TBCD; const Divisor: Comp; out Result: TBCD); inline;
  procedure BCDDivide(const Dividend: Comp; const Divisor: TBCD; out Result: TBCD); inline;
{$endif}
procedure BCDDivide(const Dividend: FmtBCDStringType; const Divisor: TBCD; out Result: TBCD); inline;

// Operators
operator =(const BCD1, BCD2: TBCD): Boolean; inline;
operator <(const BCD1, BCD2: TBCD): Boolean; inline;
operator >(const BCD1, BCD2: TBCD): Boolean; inline;
operator <=(const BCD1, BCD2: TBCD): Boolean; inline;
operator >=(const BCD1, BCD2: TBCD): Boolean; inline;
operator -(const BCD: TBCD): TBCD; inline;
operator +(const BCD1, BCD2: TBCD): TBCD; inline;
operator +(const BCD1: TBCD; const Value2: Int64): TBCD; inline;
operator +(const Value1: Int64; const BCD2: TBCD): TBCD; inline;
{$ifndef FPUNONE}
  operator +(const BCD1: TBCD; const Value2: ValReal): TBCD; inline;
  operator +(const Value1: ValReal; const BCD2: TBCD): TBCD; inline;
{$endif}
operator +(const BCD1: TBCD; const Value2: Currency): TBCD; inline;
operator +(const Value1: Currency; const BCD2: TBCD): TBCD; inline;
{$ifdef SUPPORT_COMP}
  operator +(const BCD1: TBCD; const Value2: Comp): TBCD; inline;
  operator +(const Value1: Comp; const BCD2: TBCD): TBCD; inline;
{$endif}
operator +(const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD; inline;
operator +(const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD; inline;
operator -(const BCD1, BCD2: TBCD): TBCD; inline;
operator -(const BCD1: TBCD; const Value2: Int64): TBCD; inline;
operator -(const Value1: Int64; const BCD2: TBCD): TBCD; inline;
{$ifndef FPUNONE}
  operator -(const BCD1: TBCD; const Value2: ValReal): TBCD; inline;
  operator -(const Value1: ValReal; const BCD2: TBCD): TBCD; inline;
{$endif}
operator -(const BCD1: TBCD; const Value2: Currency): TBCD; inline;
operator -(const Value1: Currency; const BCD2: TBCD): TBCD; inline;
{$ifdef SUPPORT_COMP}
  operator -(const BCD1: TBCD; const Value2: Comp): TBCD; inline;
  operator -(const Value1: Comp; const BCD2: TBCD): TBCD; inline;
{$endif}
operator -(const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD; inline;
operator -(const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD; inline;
operator *(const BCD1, BCD2: TBCD): TBCD; inline;
operator *(const BCD1: TBCD; const Value2: Int64): TBCD; inline;
operator *(const Value1: Int64; const BCD2: TBCD): TBCD; inline;
{$ifndef FPUNONE}
  operator *(const BCD1: TBCD; const Value2: ValReal): TBCD; inline;
  operator *(const Value1: ValReal; const BCD2: TBCD): TBCD; inline;
{$endif}
operator *(const BCD1: TBCD; const Value2: Currency): TBCD; inline;
operator *(const Value1: Currency; const BCD2: TBCD): TBCD; inline;
{$ifdef SUPPORT_COMP}
  operator *(const BCD1: TBCD; const Value2: Comp): TBCD; inline;
  operator *(const Value1: Comp; const BCD2: TBCD): TBCD; inline;
{$endif}
operator *(const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD; inline;
operator *(const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD; inline;
operator /(const BCD1, BCD2: TBCD): TBCD; inline;
operator /(const BCD1: TBCD; const Value2: Int64): TBCD; inline;
operator /(const Value1: Int64; const BCD2: TBCD): TBCD; inline;
{$ifndef FPUNONE}
  operator /(const BCD1: TBCD; const Value2: ValReal): TBCD; inline;
  operator /(const Value1: ValReal; const BCD2: TBCD): TBCD; inline;
{$endif}
operator /(const BCD1: TBCD; const Value2: Currency): TBCD; inline;
operator /(const Value1: Currency; const BCD2: TBCD): TBCD; inline;
{$ifdef SUPPORT_COMP}
  operator /(const BCD1: TBCD; const Value2: Comp): TBCD; inline;
  operator /(const Value1: Comp; const BCD2: TBCD): TBCD; inline;
{$endif}
operator /(const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD; inline;
operator /(const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD; inline;
operator :=(const Value: QWord): TBCD; inline;
operator :=(const BCD: TBCD): QWord; inline;
operator :=(const Value: int64): TBCD; inline;
operator :=(const BCD: TBCD): Int64; inline;
{$ifndef FPUNONE}
  operator :=(const Value: Double): TBCD; inline;
  operator :=(const BCD: TBCD): Double; inline;
{$endif}
operator :=(const Value: Currency): TBCD; inline;
operator :=(const BCD: TBCD): Currency; inline;
{$ifdef SUPPORT_COMP}
  operator :=(const Value: Comp): TBCD; inline;
  operator :=(const BCD: TBCD): Comp; inline;
{$endif}
operator :=(const Value: ShortString): TBCD; inline;
operator :=(const BCD: TBCD): ShortString; inline;
operator :=(const Value: AnsiString): TBCD; inline;
operator :=(const BCD: TBCD): AnsiString; inline;
{$ENDIF SKIP}

function NullBCD: TBCD;
function OneBCD: TBCD;

implementation

uses Classes, Math;

type
  // Unpacked BCd for calculation
  // Lets number is 0000ABC.defgh00000, then
  // Digits[-2..0] = ABC, Digits[1..5] = defgh, all other digits is '0'
  // Precision is count(ABC) + count(defgh) = 8
  // LowDigits is -count(ABC) + 1 = -2
  // HighDigits is count(defgh) = 5
  // Digits is 64 digits before point + 63 digits after point
  TUnpackedBCD = record
  private
    LowDigits: Integer; // max 1 (1 = no whole part)
    HighDigits: Integer; // min 0 (0 = no frac part)
    Digits: array[-TBCD.CMaxPrecision + 1..TBCD.CMaxPrecision - 1] of Byte; // of 0..9
    Negative: Boolean;
  public
    procedure Init(ALowDigits, AHighDigits: Integer; ANegative: Boolean = False);
    procedure ErrorIfNotNormalized(const ErrorInRoutine: string);
    function Normalize: Boolean; // Return false, if not fit into BCD
    function Precision: Integer; inline;
    function Power10(Power: Integer): Boolean; // X*10^Power, False if very big
    function Round: Boolean; // Return false, if not fit into BCD
    procedure SkipTrailingZero;
    property Places: Integer read HighDigits write HighDigits;
  end;

  procedure FillZeroBytes(Addr: Pointer; Count: Integer);
  begin
    FillChar(Addr^, Count, 0);
  end;

  function NullBCD: TBCD;
  begin
    FillZeroBytes(@Result, SizeOf(Result));
  end;

  function OneBCD: TBCD;
  begin
    FillZeroBytes(@Result, SizeOf(Result));
    Result.FPrecision := 1;
    Result.FData[Low(Result.FData)] := $10;
  end;


  { TBCD }

  function TBCD.GetSignSpecialPlaces: Byte;
  begin
    Result := Byte(FInfo);
  end;

  procedure TBCD.SetNegative(const AValue: Boolean);
  begin
    FInfo.Negative := AValue;
  end;

  procedure TUnpackedBCD.Init(ALowDigits, AHighDigits: Integer;
    ANegative: Boolean = False);
  begin
    LowDigits := ALowDigits;
    HighDigits := AHighDigits;
    Negative := ANegative;
    FillZeroBytes(@Digits, SizeOf(Digits));
  end;

  // For inlining it's musb be above first using!
  function TUnpackedBCD.Precision: Integer; inline;
  begin
    Result := HighDigits - LowDigits + 1;
  end;

  procedure TUnpackedBCD.ErrorIfNotNormalized(const ErrorInRoutine: string);

    procedure DoError(const ErrorInRoutine: string);
    begin
      raise EBCDOverflowException.Create('in ' + ErrorInRoutine);
    end;

  begin
    if not Normalize then
      DoError(ErrorInRoutine)
  end;

  function TUnpackedBCD.Normalize: Boolean;
  var
    NeedRound: Boolean = False;
  begin
    // Skip leading zero
    while (LowDigits <= 0) and (Digits[LowDigits] = 0) do
      Inc(LowDigits);
    SkipTrailingZero;
    if HighDigits > High(Digits) then // very long fraction - do trunc
    begin
      NeedRound := True;
      HighDigits := High(Digits);
    end;
    if Precision > TBCD.CMaxPrecision then // very long - do trunc
    begin
      NeedRound := True;
      Dec(HighDigits, TBCD.CMaxPrecision - Precision);
      if HighDigits < 0 then
        Exit(False);
    end;
    if NeedRound and not Round then
      Exit(False);
    Result := True;
  end;

  function TUnpackedBCD.Power10(Power: Integer): Boolean;
  var
    CountToShift: Integer;
  begin
    CountToShift := Precision;
    if Power < 0 then // shift right
    begin
      Power := -Power;
      Inc(HighDigits, Power);
      if HighDigits > High(Digits) then // Very small, ignore rest
      begin
        Dec(CountToShift, HighDigits - High(Digits));
        HighDigits := High(Digits);
      end;
      Move(Digits[LowDigits], Digits[LowDigits + Power], CountToShift);
      FillZeroBytes(@Digits[LowDigits], CountToShift);
      Inc(LowDigits, Power);
      if LowDigits > 1 then
        LowDigits := 1;
    end
    else // shift left
    begin
      Dec(LowDigits, Power);
      if LowDigits < Low(Digits) then
        Exit(False); // Not fit
      Move(Digits[LowDigits + Power], Digits[LowDigits], CountToShift);
      FillZeroBytes(@Digits[LowDigits + Power], CountToShift);
      Dec(HighDigits, Power);
      if HighDigits < 0 then
        HighDigits := 0;
    end;
    Result := True;
  end;

  function TUnpackedBCD.Round: Boolean;
  var
    Overflow: Boolean;
    i: Integer;
  begin
    i := HighDigits;
    Overflow := (Digits[i + 1] >= 5);
    while (i >= LowDigits) and Overflow do
    begin
      Overflow := (Digits[i] = 9);
      if Overflow then
        Digits[i] := 0
      else
        Inc(Digits[i]);
      Inc(i);
    end;
    SkipTrailingZero;
    if Overflow then
    begin
      Dec(LowDigits);
      if Precision > TBCD.CMaxPrecision then
        Exit(False);
      Digits[LowDigits] := 1;
    end;
    Result := True;
  end;

  procedure TUnpackedBCD.SkipTrailingZero;
  begin
    while (HighDigits > 1) and (Digits[HighDigits] = 0) do
      Dec(HighDigits);
  end;

  {$IFNDEF SKIP}

  { TFMTBcdFactory }
  TFMTBcdFactory = class(TPublishableVariantType)
  protected
    function GetInstance(const V: TVarData): TObject; override;
  public
    procedure BinaryOp(var Left: TVarData; const Right: TVarData; const Operation: TVarOp); override;
    procedure Clear(var V: TVarData); override;
    procedure Copy(var Dest: TVarData; const Source: TVarData; const Indirect: Boolean); override;
    function CompareOp(const Left, Right: TVarData; const Operation: TVarOp): Boolean; override;
    procedure Compare(const Left, Right: TVarData; var Relationship: TVarCompareResult); override;
    procedure Cast(var Dest: TVarData; const Source: TVarData); override;
    procedure CastTo(var Dest: TVarData; const Source: TVarData; const aVarType: TVarType); override;
  end;

  { TFMTBcdVarData }
  TFMTBcdVarData = class(TPersistent)
  private
    FBcd: TBCD;
  public
    constructor Create;
    constructor Create(const BCD: TBCD);
    property BCD: TBCD read FBcd write FBcd;
  end;

var
  GFMTBcdFactory: TFMTBcdFactory = nil;
{$ENDIF SKIP}


function BCDPrecision(const BCD: TBCD): Word;
begin
  Result := BCD.Precision;
end;

function BCDScale(const BCD: TBCD): Word;
begin
  Result := BCD.Places;
end;

function IsBCDNegative(const BCD: TBCD): Boolean;
begin
  Result := BCD.Negative;
end;

procedure BCDNegate(var BCD: TBCD);
begin
  BCD.Negative := not BCD.Negative;
end;

function UnpackBCD(const BCD: TBCD): TUnpackedBCD;
var
  iDigits, iData: Integer;
  B: Byte;
begin
  Result.Init(BCD.Places - BCD.Precision + 1, BCD.Places, BCD.Negative);
  iDigits := Result.LowDigits;
  iData := 0;
  // loop by full two digits, may be not include odd last
  while iDigits < Result.HighDigits do
  begin
    B := BCD.FData[iData];
    Inc(iData);
    Result.Digits[iDigits] := B shr 4;
    Inc(iDigits);
    Result.Digits[iDigits] := B and $0F;
    Inc(iDigits);
  end;
  if iDigits <= Result.HighDigits then
    Result.Digits[iDigits] := BCD.FData[iData] shr 4;
end;

function PackBCD(const U: TUnpackedBCD): TBCD;
var
  iDigits, iData: Integer;
begin
  Result.FPrecision := U.Precision;
  Result.FInfo.Places := U.Places;
  Result.FInfo.Negative := U.Negative;
  FillZeroBytes(@Result.FData, SizeOf(Result.FData));
  iDigits := U.LowDigits;
  iData := 0;
  // loop by full two digits, may be not include odd last
  while iDigits < U.HighDigits do
  begin
    Result.FData[iData] := U.Digits[iDigits] shl 4 + U.Digits[iDigits + 1];
    Inc(iDigits, 2);
    Inc(iData);
  end;
  if iDigits <= U.HighDigits then
    Result.FData[iData] := U.Digits[iDigits] shl 4;
end;

function NormalizeBCD(const InBCD: TBCD; out OutBCD: TBCD; const MaxPrecision,
  MaxPlaces: Integer): Boolean;

  function Repack(const InBCD: TBCD; const MaxPrecision, MaxPlaces: Integer): TBCD;
  var
    U: TUnpackedBCD;
  begin
    U := UnpackBCD(InBCD);
    if U.Places > MaxPlaces then
    begin
      U.Places := MaxPlaces;
      U.Round;
    end;
    if U.Precision > MaxPrecision then
      U.LowDigits := U.HighDigits - MaxPrecision + 1;
    U.ErrorIfNotNormalized('NormalizeBCD');
    Result := PackBCD(U);
  end;

var
  LMaxPrecision, LMaxPlaces: Integer;
begin
  Assert(MaxPrecision in [0..TBCD.CMaxPrecision],
    'Parameter MaxPrecision in NormalizeBCD is out of range');
  Assert(MaxPlaces in [0..TBCD.CMaxPrecision - 1],
    'Parameter MaxPlaces in NormalizeBCD is out of range');
  LMaxPrecision := EnsureRange(MaxPrecision, 0, TBCD.CMaxPrecision);
  LMaxPlaces := EnsureRange(MaxPlaces, 0, TBCD.CMaxPrecision - 1);
  if (InBCD.Places <= LMaxPlaces) and (InBCD.Precision <= LMaxPrecision) then
  begin
    OutBCD := InBCD;
    Result := True;
  end
  else
  begin
    OutBCD := Repack(InBcd, LMaxPrecision, LMaxPlaces);
    Result := False;
  end;
end;

{$IFNDEF SKIP}
procedure VarFmtBCDCreate(out Dest: Variant; const BCD: TBCD);
begin
  VarClear(Dest);
  TVarData(Dest).VType := GFMTBcdFactory.VarType;
  TVarData(Dest).VPointer := TFMTBcdVarData.Create(BCD);
end;

function VarFmtBCDCreate: Variant;
begin
  VarFmtBCDCreate(Result, NullBCD);
end;

function VarFmtBCDCreate(const Value: FmtBCDStringType;
  Precision, Scale: Word): Variant;
begin
  VarFmtBCDCreate(Result, StrToBCD(Value));
end;

function VarFmtBCDCreate(const Value: ValReal; Precision: Word;
  Scale: Word): Variant;
begin
  VarFmtBCDCreate(Result, DoubleToBCD(Value));
end;

function VarFmtBCDCreate(const BCD: TBCD): Variant;
begin
  VarFmtBCDCreate(Result, BCD);
end;

function VarIsFmtBCD(const Value: Variant): Boolean;
begin
  Result := (TVarData(Value).VType = GFMTBcdFactory.VarType);
end;

function VarFmtBCD: TVarType;
begin
  Result := GFMTBcdFactory.VarType;
end;
{$ENDIF SKIP}

function StrToBCD(const Value: FmtBCDStringType): TBCD;
begin
  Result := StrToBCD(Value, DefaultFormatSettings);
end;

function StrToBCD(const Value: FmtBCDStringType;
  const Format: TFormatSettings): TBCD;

  procedure Error;
  begin
    raise EBCDOverflowException.Create('in StrToBCD');
  end;

begin
  if not TryStrToBCD(Value, Result, Format) then
    Error;
end;

function TryStrToBCD(const Value: FmtBCDStringType; out BCD: TBCD): Boolean;
begin
  Result := TryStrToBCD(Value, BCD, DefaultFormatSettings);
end;

function TryStrToBCD(const Value: FmtBCDStringType; out BCD: TBCD;
  const Format: TFormatSettings): Boolean;
// Full format '  +#t#d#E+#  ', where '+' = sign, 't' - thousand separator,
// 'd' - decimal separator, '#' - one or more digits, spaces is spaces
// Accept "#.E#", ".#E#", ".E#", "E#", 'E'
type
  TSign = (siUnassigned, siPlus, siMinus);

  function SetSign(var Sign: TSign; Ch: Char): Boolean;
  begin
    Result := (Sign = siUnassigned);
    if Result then
      if Ch = '+' then
        Sign := siPlus
      else
        Sign := siMinus;
  end;

var
  iValue: Integer;

  function SkipTrailingSpaces: Boolean;
  begin
    while (iValue <= Length(Value)) and (Value[iValue] = ' ') do
      Inc(iValue);
    Result := (iValue > Length(Value));
  end;

var
  Exp: Integer;

  function ScanExp: Boolean;
  var
    Ch: Char;
    SignExp: TSign;
  begin
    SignExp := siUnassigned;
    while iValue <= Length(Value) do
    begin
      Ch := Value[iValue];
      Inc(iValue);
      case Ch of
        '0'..'9':
          begin
            if Exp >= TBCD.CMaxPrecision then
              Exit(False); // Very big
            Exp := Exp * 10 + Ord(Ch) - Ord('0');
          end;
        '+', '-':
          if not SetSign(SignExp, Ch) then
            Exit(False); // repeat sign
        ' ':
          Exit(SkipTrailingSpaces);
      else
        Exit(False);
      end;
    end;
    if SignExp = siMinus then
      Exp := -Exp;
    Result := True;
  end;

var
  U: TUnpackedBCD;

  function ScanFrac: Boolean;
  var
    Ch: Char;
  begin
    while iValue <= Length(Value) do
    begin
      Ch := Value[iValue];
      Inc(iValue);
      case Ch of
        'e', 'E':
          Exit(ScanExp);
        '0'..'9':
          if U.HighDigits < High(U.Digits) then
          begin
            Inc(U.HighDigits);
            U.Digits[U.HighDigits] := Ord(Ch) - Ord('0');
          end;
        ' ':
          Exit(SkipTrailingSpaces);
      else
        Exit(False);
      end;
    end;
    Result := True;
  end;

  function ScanInt: Boolean;
  var
    Sign: TSign;
    Ch: Char;
  begin
    Sign := siUnassigned;
    while iValue <= Length(Value) do
    begin
      Ch := Value[iValue];
      Inc(iValue);
      if Ch = Format.ThousandSeparator then
        Continue; // Skip it
      if Ch = Format.DecimalSeparator then
        Exit(ScanFrac);
      case Ch of
        '0'..'9':
          begin
            if not U.Power10(1) then // Frac zero in this
              Exit(False); // Very big
            U.Digits[0] := Ord(Ch) - Ord('0');
          end;
        '+', '-':
          if not SetSign(Sign, Ch) then
            Exit(False); // repeat sign
        'e', 'E':
          Exit(ScanExp);
        ' ':
          Exit(SkipTrailingSpaces);
      else
        Exit(False);
      end;
    end;
    U.Negative := (Sign = siMinus);
    Result := True;
  end;

begin
  BCD := NullBCD; // for False result
  // Skip leading spaces
  iValue := 1;
  while (iValue <= Length(Value)) and (Value[iValue] = ' ') do
    Inc(iValue);
  if iValue = Length(Value) then
    Exit(False); // Empty Value
  U.Init(1, 0); // Null unpacked BCD
  Exp := 0;
  if not ScanInt then
    Exit(False);
  if Exp <> 0 then
    if not U.Power10(Exp) then
      Exit(False);
  Result := U.Normalize;
  if Result then
    BCD := PackBCD(U);
end;

{$IFNDEF SKIP}

function DoubleToBCD(const Value: ValReal): TBCD;
var
  S: string[30];
  F: TFormatSettings;
begin
  System.Str(Value:25, S);
  f.DecimalSeparator := '.';
  f.ThousandSeparator := #0;
  Result := StrToBCD(S, F);
end;

procedure DoubleToBCD(const Value: ValReal; out Result: TBCD);
begin
  DoubleToBCD(Value, Result);
end;

function IntegerToBCD(const Value: Int64): TBCD;
begin
  // DoIt
end;

function VariantToBCD(const VarSrc: TVarData): TBCD;
begin
  // Doit
end;

function VarToBCD(const Value: Variant): TBCD;
begin
  Result := VariantToBCD(TVarData(Value));
end;

function CurrToBCD(const Value: Currency; out BCD: TBCD; Precision: Integer;
  Decimals: Integer): Boolean;
var
  // this works under the assumption that a Currency is an int64 except for scale of 10000
  i: Int64 absolute Value;
begin
  BCD := IntegerToBCD(i);
  BCD.FInfo.Places := 4;
end;

function BCDToStr(const BCD: TBCD): FmtBCDStringType;
begin
  Result := BCDToStr(BCD, DefaultFormatSettings);
end;

function BCDToStr(const BCD: TBCD; const Format: TFormatSettings): FmtBCDStringType;
begin
  // Doit
end;

function BCDToDouble(const BCD: TBCD): ValReal;
begin
  // Doit
end;

function BCDToInteger(const BCD: TBCD; Truncate: Boolean): Int64;
begin
  // Doit
end;

function BCDToCurr(const BCD: TBCD; out OutResult: currency): Boolean;
begin
  // Doit
end;

function BCDToStrF(const BCD: TBCD; Format: TFloatFormat; const Precision,
  Digits: Integer): FmtBCDStringType;
begin
  // Doit
end;

function FormatBCD(const Format: string; const BCD: TBCD): FmtBCDStringType;
begin
  // Doit
end;

function BCDCompare(const BCD1, BCD2: TBCD): Integer;
begin
  // Doit
end;

function CurrToBCD(const Value: Currency): TBCD;
begin
  CurrToBCD(Value, Result);
end;

procedure BCDAdd(const BCD1, BCD2: TBCD; out Result: TBCD);
begin
  // Doit
end;

procedure BCDAdd(const BCD1: TBCD; const Int2: Int64; out Result: TBCD);
begin
  // Doit
end;

procedure BCDAdd(const Int1: Int64; const BCD2: TBCD; out Result: TBCD);
begin
  BCDAdd(BCD2, Int1, Result);
end;

procedure BCDAdd(const BCD1: TBCD; const Double2: ValReal; out Result: TBCD);
begin
  BCDAdd(BCD1, DoubleToBCD(Double2), Result);
end;

procedure BCDAdd(const Double1: ValReal; const BCD2: TBCD; out Result: TBCD);
begin
  BCDAdd(DoubleToBCD(Double1), BCD2, Result);
end;

procedure BCDAdd(const BCD1: TBCD; const Curr2: Currency; out Result: TBCD);
begin
  BCDAdd(BCD1, CurrToBCD(Curr2), Result);
end;

procedure BCDAdd(const Curr1: Currency; const BCD2: TBCD; out Result: TBCD);
begin
  BCDAdd(CurrToBCD(Curr1), BCD2, Result);
end;

procedure BCDAdd(const BCD1: TBCD; const Str2: FmtBCDStringType; out
  Result: TBCD);
begin
  BCDAdd(BCD1, StrToBCD(Str2), Result);
end;

procedure BCDAdd(const Str1: FmtBCDStringType; const BCD2: TBCD; out Result: TBCD);
begin
  BCDAdd(StrToBCD(Str1), BCD2, Result);
end;

procedure BCDAdd(const Str1, Str2: FmtBCDStringType; out Result: TBCD);
begin
  BCDAdd(StrToBCD(Str1), StrToBCD(Str2), Result);
end;

procedure BCDSubtract(const BCD1, BCD2: TBCD; out Result: TBCD);
begin
  // Doit
end;

procedure BCDSubtract(const BCD1: TBCD; const Int2: Int64; out Result: TBCD);
begin
  // Doit
end;

procedure BCDSubtract(const Int1: Int64; const BCD2: TBCD; out Result: TBCD);
begin
  BCDSubtract(BCD2, Int1, Result);
  BCDNegate(Result);
end;

procedure BCDSubtract(const BCD1: TBCD; const Double2: ValReal; out Result: TBCD);
begin
  BCDSubtract(BCD1, DoubleToBCD(Double2), Result);
end;

procedure BCDSubtract(const Double1: ValReal; const BCD2: TBCD; out Result: TBCD);
begin
  BCDSubtract(DoubleToBCD(Double1), BCD2, Result);
end;

procedure BCDSubtract(const BCD1: TBCD; const Curr2: Currency; out Result: TBCD);
begin
  BCDSubtract(BCD1, CurrToBCD(Curr2), Result);
end;

procedure BCDSubtract(const Curr1: Currency; const BCD2: TBCD; out Result: TBCD);
begin
  BCDSubtract(CurrToBCD(Curr1), BCD2, Result);
end;

procedure BCDSubtract(const BCD1: TBCD; const Str2: FmtBCDStringType; out Result: TBCD);
begin
  BCDSubtract(BCD1, StrToBCD(Str2), Result);
end;

procedure BCDSubtract(const Str1: FmtBCDStringType; const BCD2: TBCD; out Result: TBCD);
begin
  BCDSubtract(StrToBCD(Str1), BCD2, Result);
end;

procedure BCDSubtract(const Str1, Str2: FmtBCDStringType; out Result: TBCD);
begin
  BCDSubtract(StrToBCD(Str1), StrToBCD(Str2), Result);
end;

procedure BCDMultiply(const BCD1, BCD2: TBCD; out Result: TBCD);
begin
  // Doit
end;

procedure BCDMultiply(const BCD1: TBCD; const Double2: ValReal; out Result: TBCD);
begin
  BCDMultiply(BCD1, DoubleToBCD(Double2), Result);
end;

procedure BCDMultiply(const BCD1: TBCD; const Str2: FmtBCDStringType; out Result: TBCD);
begin
  BCDMultiply(BCD1, StrToBCD(Str2), Result);
end;

procedure BCDMultiply(const Str1, Str2: FmtBCDStringType; out Result: TBCD);
begin
  BCDMultiply(StrToBCD(Str1), StrToBCD(Str2), Result);
end;

procedure BCDMultiply(const BCD1: TBCD; const Int2: Int64; out Result: TBCD);
begin
  // Doit
end;

procedure BCDMultiply(const Int1: Int64; const BCD2: TBCD; out Result: TBCD);
begin
  BCDMultiply(BCD2, Int1, Result);
end;

procedure BCDMultiply(const Double1: ValReal; const BCD2: TBCD; out Result: TBCD);
begin
  BCDMultiply(DoubleToBCD(Double1), BCD2, Result);
end;

procedure BCDMultiply(const BCD1: TBCD; const Curr2: Currency; out Result: TBCD);
begin
  BCDMultiply(BCD1, CurrToBCD(Curr2), Result);
end;

procedure BCDMultiply(const Curr1: Currency; const BCD2: TBCD; out Result: TBCD);
begin
  BCDMultiply(CurrToBCD(Curr1), BCD2, Result);
end;

procedure BCDMultiply(const Str1: FmtBCDStringType; const BCD2: TBCD; out Result: TBCD);
begin
  BCDMultiply(StrToBCD(Str1), BCD2, Result);
end;

procedure BCDDivide(const Dividend, Divisor: TBCD; out Result: TBCD);
begin
  // Doit
end;

procedure BCDDivide(const Dividend: TBCD; const Divisor: ValReal; out Result: TBCD);
begin
  BCDDivide(Dividend, DoubleToBCD(Divisor), Result);
end;

procedure BCDDivide(const Dividend: TBCD; const Divisor: FmtBCDStringType;
  out Result: TBCD);
begin
  BCDDivide(Dividend, StrToBCD(Divisor), Result);
end;

procedure BCDDivide(const Dividend, Divisor: FmtBCDStringType; out Result: TBCD);
begin
  BCDDivide(StrToBCD(Dividend), StrToBCD(Divisor), Result);
end;

procedure BCDDivide(const Dividend: TBCD; const Divisor: Int64; out Result: TBCD);
begin
  BCDDivide(Dividend, IntegerToBCD(Divisor), Result);
end;

procedure BCDDivide(const Dividend: Int64; const Divisor: TBCD; out Result: TBCD);
begin
  BCDDivide(IntegerToBCD(Dividend), Divisor, Result);
end;

procedure BCDDivide(const Dividend: ValReal; const Divisor: TBCD; out Result: TBCD);
begin
  BCDDivide(DoubleToBCD(Dividend), Divisor, Result);
end;

procedure BCDDivide(const Dividend: TBCD; const Divisor: Currency; out Result: TBCD);
begin
  BCDDivide(Dividend, CurrToBCD(Divisor), Result);
end;

procedure BCDDivide(const Dividend: Currency; const Divisor: TBCD; out Result: TBCD);
begin
  BCDDivide(CurrToBCD(Dividend), Divisor, Result);
end;

procedure BCDDivide(const Dividend: FmtBCDStringType; const Divisor: TBCD;
  out Result: TBCD);
begin
  BCDDivide(StrToBCD(Dividend), Divisor, Result);
end;

operator = (const BCD1, BCD2: TBCD): Boolean;
begin
  Result := (BCDCompare(BCD1, BCD2) = 0);
end;

operator < (const BCD1, BCD2: TBCD): Boolean;
begin
  Result := (BCDCompare(BCD1, BCD2) < 0);
end;

operator > (const BCD1, BCD2: TBCD): Boolean;
begin
  Result := (BCDCompare(BCD1, BCD2) > 0);
end;

operator <= (const BCD1, BCD2: TBCD): Boolean;
begin
  Result := (BCDCompare(BCD1, BCD2) <= 0);
end;

operator >= (const BCD1, BCD2: TBCD): Boolean;
begin
  Result := (BCDCompare(BCD1, BCD2) >= 0);
end;

operator - (const BCD: TBCD): TBCD;
begin
  Result := BCD;
  BCDNegate(Result);
end;

operator + (const BCD1, BCD2: TBCD): TBCD;
begin
  BCDAdd(BCD1, BCD2, Result);
end;

operator + (const BCD1: TBCD; const Value2: Int64): TBCD;
begin
  BCDAdd(BCD1, Value2, Result);
end;

operator + (const Value1: Int64; const BCD2: TBCD): TBCD;
begin
  BCDAdd(Value1, BCD2, Result);
end;

operator + (const BCD1: TBCD; const Value2: ValReal): TBCD;
begin
  BCDAdd(BCD1, Value2, Result);
end;

operator + (const Value1: ValReal; const BCD2: TBCD): TBCD;
begin
  BCDAdd(Value1, BCD2, Result);
end;

operator + (const BCD1: TBCD; const Value2: Currency): TBCD;
begin
  BCDAdd(BCD1, Value2, Result);
end;

operator + (const Value1: Currency; const BCD2: TBCD): TBCD;
begin
  BCDAdd(Value1, BCD2, Result);
end;

operator + (const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD;
begin
  BCDAdd(BCD1, Value2, Result);
end;

operator + (const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD;
begin
  BCDAdd(Value1, BCD2, Result);
end;

operator - (const BCD1, BCD2: TBCD): TBCD;
begin
  BCDSubtract(BCD1, BCD2, Result);
end;

operator - (const BCD1: TBCD; const Value2: Int64): TBCD;
begin
  BCDSubtract(BCD1, Value2, Result);
end;

operator - (const Value1: Int64; const BCD2: TBCD): TBCD;
begin
  BCDSubtract(Value1, BCD2, Result);
end;

operator - (const BCD1: TBCD; const Value2: ValReal): TBCD;
begin
  BCDSubtract(BCD1, Value2, Result);
end;

operator - (const Value1: ValReal; const BCD2: TBCD): TBCD;
begin
  BCDSubtract(Value1, BCD2, Result);
end;

operator - (const BCD1: TBCD; const Value2: Currency): TBCD;
begin
  BCDSubtract(BCD1, Value2, Result);
end;

operator - (const Value1: Currency; const BCD2: TBCD): TBCD;
begin
  BCDSubtract(Value1, BCD2, Result);
end;

operator - (const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD;
begin
  BCDSubtract(BCD1, Value2, Result);
end;

operator - (const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD;
begin
  BCDSubtract(Value1, BCD2, Result);
end;

operator * (const BCD1, BCD2: TBCD): TBCD;
begin
  BCDMultiply(BCD1, BCD2, Result);
end;

operator * (const BCD1: TBCD; const Value2: Int64): TBCD;
begin
  BCDMultiply(BCD1, Value2, Result);
end;

operator * (const Value1: Int64; const BCD2: TBCD): TBCD;
begin
  BCDMultiply(Value1, BCD2, Result);
end;

operator * (const BCD1: TBCD; const Value2: ValReal): TBCD;
begin
  BCDMultiply(BCD1, Value2, Result);
end;

operator * (const Value1: ValReal; const BCD2: TBCD): TBCD;
begin
  BCDMultiply(Value1, BCD2, Result);
end;

operator * (const BCD1: TBCD; const Value2: Currency): TBCD;
begin
  BCDMultiply(BCD1, Value2, Result);
end;

operator * (const Value1: Currency; const BCD2: TBCD): TBCD;
begin
  BCDMultiply(Value1, BCD2, Result);
end;

operator * (const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD;
begin
  BCDMultiply(BCD1, Value2, Result);
end;

operator * (const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD;
begin
  BCDMultiply(Value1, BCD2, Result);
end;

operator / (const BCD1, BCD2: TBCD): TBCD;
begin
  BCDDivide(BCD1, BCD2, Result);
end;

operator / (const BCD1: TBCD; const Value2: Int64): TBCD;
begin
  BCDDivide(BCD1, Value2, Result);
end;

operator / (const Value1: Int64; const BCD2: TBCD): TBCD;
begin
  BCDDivide(Value1, BCD2, Result);
end;

operator / (const BCD1: TBCD; const Value2: ValReal): TBCD;
begin
  BCDDivide(BCD1, Value2, Result);
end;

operator / (const Value1: ValReal; const BCD2: TBCD): TBCD;
begin
  BCDDivide(Value1, BCD2, Result);
end;

operator / (const BCD1: TBCD; const Value2: Currency): TBCD;
begin
  BCDDivide(BCD1, Value2, Result);
end;

operator / (const Value1: Currency; const BCD2: TBCD): TBCD;
begin
  BCDDivide(Value1, BCD2, Result);
end;

operator / (const BCD1: TBCD; const Value2: FmtBCDStringType): TBCD;
begin
  BCDDivide(BCD1, Value2, Result);
end;

operator / (const Value1: FmtBCDStringType; const BCD2: TBCD): TBCD;
begin
  BCDDivide(Value1, BCD2, Result);
end;

operator := (const Value: QWord): TBCD;
begin
  Result := IntegerToBCD(Value);
end;

operator := (const BCD: TBCD): QWord;
begin
  Result := BCDToInteger(BCD);
end;

operator := (const Value: Int64): TBCD;
begin
  Result := IntegerToBCD(Value);
end;

operator := (const BCD: TBCD): Int64;
begin
  Result := BCDToInteger(BCD);
end;

operator := (const Value: Double): TBCD;
begin
  Result := DoubleToBCD(Value);
end;

operator := (const BCD: TBCD): Double;
begin
  Result := BCDToDouble(BCD);
end;

operator := (const Value: Currency): TBCD;
begin
  Result := CurrToBCD(Value);
end;

operator := (const BCD: TBCD): Currency;
begin
  BCDToCurr(BCD, Result);
end;

operator := (const Value: ShortString): TBCD;
begin
  Result := StrToBCD(Value);
end;

operator := (const BCD: TBCD): ShortString;
begin
  Result := BCDToStr(BCD);
end;

operator := (const Value: AnsiString): TBCD;
begin
  Result := StrToBCD(Value);
end;

operator := (const BCD: TBCD): AnsiString;
begin
  Result := BCDToStr(BCD);
end;

{ TFMTBcdVarData }

constructor TFMTBcdVarData.Create(const BCD: TBCD);
begin

end;

constructor TFMTBcdVarData.Create;
begin

end;

{ TFMTBcdFactory }

procedure TFMTBcdFactory.BinaryOp(var Left: TVarData; const Right: TVarData;
  const Operation: TVarOp);
begin
  inherited BinaryOp(Left, Right, Operation);
end;

procedure TFMTBcdFactory.Cast(var Dest: TVarData; const Source: TVarData);
begin
  inherited Cast(Dest, Source);
end;

procedure TFMTBcdFactory.CastTo(var Dest: TVarData; const Source: TVarData;
  const aVarType: TVarType);
begin
  inherited CastTo(Dest, Source, aVarType);
end;

procedure TFMTBcdFactory.Clear(var V: TVarData);
begin

end;

procedure TFMTBcdFactory.Compare(const Left, Right: TVarData;
  var Relationship: TVarCompareResult);
begin
  inherited Compare(Left, Right, Relationship);
end;

function TFMTBcdFactory.CompareOp(const Left, Right: TVarData;
  const Operation: TVarOp): Boolean;
begin
  Result := inherited CompareOp(Left, Right, Operation);
end;

procedure TFMTBcdFactory.Copy(var Dest: TVarData; const Source: TVarData;
  const Indirect: Boolean);
begin

end;

function TFMTBcdFactory.GetInstance(const V: TVarData): TObject;
begin

end;
{$ENDIF SKIP}


initialization
  //GFMTBcdFactory := TFMTBcdFactory.Create;
finalization
  //FreeAndNil(GFMTBcdFactory);
end.

